import express, {Request, Response} from 'express';
import { join } from 'path';
import { query } from './lib/db';
import { v4 as uuidv4 } from 'uuid';
import { ingest, ingestSingle } from './ingester';
import {RSSParser} from './parsers/rss';
import bodyParser from 'body-parser';


const app = express();

app.use(bodyParser.json());
app.use(bodyParser.urlencoded({ extended: true}));
app.use(express.static(join(__dirname, '..', 'html')));
app.use((req, res, next) => {
  console.log(`${req.method} ${req.path}`);
  next();
});

type WrappedApiHandler = (req: Request, res: Response) => Promise<any>;

function apiWrapper(method: 'get' | 'post' | 'delete', endpoint: string, fn: WrappedApiHandler, view?: (args: any) => string) : void {
  app[method](endpoint, async(req, res) => {
    try {
      const output = await fn(req, res);
      if(req.headers['content-type'] === 'application/json') {
        res.json(output);
        return;
      }

      if(view) {
        const viewOutput = view(output);
        if(viewOutput.length) {
          res.send(viewOutput);
          return;
        }
      }

      res.status(204);
    }
    catch(e) {
      console.error(e);
      res.status(500);
    }
    finally {
      res.end();
    }
  });
}

function apiGet(endpoint: string, fn: WrappedApiHandler, view?: (arr: any) => string): void {
  apiWrapper('get', endpoint, fn, view);
}

function apiPost(endpoint: string, fn: WrappedApiHandler, view?: (arr: any) => string): void {
  apiWrapper('post', endpoint, fn, view);
}
function apiDelete(endpoint: string, fn: WrappedApiHandler, view?: (arr: any) => string): void {
  apiWrapper('delete', endpoint, fn, view);
}

apiPost('/feeds', async (req, res): Promise<any> => {
  // get info about the feed
  const url = req.body.link;

  const rss = new RSSParser();
  const feedData = await rss.parse(url);

  const title = feedData.title;
  const id = uuidv4();

  // ingest teh feed, 

  query.addFeed.run(id, title, url);

  await ingestSingle(id);

  res.setHeader('HX-Trigger', 'newFeed');

  return {
    id,
    title,
    url
  }
});

apiGet('/feeds', async (req, res): Promise<any> => {
  const feeds = query.getFeedList.all();
  // get unread counts
  const unread_count = query.getUnreadCountForAll.all();

  const feedsWithUnread = feeds.map(feed => {
    const unread = unread_count.filter(i => i.feed_id === feed.id);
    return {
      id: feed.id,
      title: feed.title,
      link: feed.link,
      unread: (unread && unread.length) ? unread[0].unread : 0
    }
  });

  return { 
    feeds: feedsWithUnread
  }
}, (output: any): string => {
  const feeds = output.feeds;
  return `<ul class="list">${feeds.map((feed: any, idx: number) =>{ 
    const display = feed.unread ? `(${feed.unread})` : '';
    const first = idx === 0;

    return `<li><a href="#" class="${first ? 'active' : ''} ${feed.unread ? 'unread' : ''}" data-actions="activate" hx-get="/feeds/${feed.id}/items" hx-trigger="${first ? 'load,' : ''}click" hx-target="#feed-item-list" data-feed-id="${feed.id}">${feed.title} 
    <span class="unread-count">${display}</span>
    </a></li>`
  }).join("\n")}</ul>`;
});

function reasonable(date: Date): string {
  const month = date.getMonth() < 10 ? `0${date.getMonth()}` : date.getMonth();
  const day = date.getDate() < 10 ? `0${date.getDate()}` : date.getDate();
  return `${date.getFullYear()}-${month}-${day}`;
}

apiGet('/feeds/:feed_id/items', async (req, res): Promise<any> => {
  const id = req.params.feed_id;

  return query.getFeedsFor.all(id, 0);
}, (feedData: any): string => {
  return `
  ${feedData.map((item: any, index: number) => {
    const read = !!item.read_at;
    return `<li>
      <a href="#" class="${index === 0 ? 'active': ''} ${read ? '': 'unread'}" data-actions="activate" hx-get="/feeds/${item.feed_id}/items/${item.id}" hx-trigger="${index === 0 ? 'load,': ''}click" hx-target="#reading-pane" data-feed-item-id="${item.id}" data-feed-id="${item.feed_id}">${item.title}
      <span class="date">${reasonable(new Date(item.pub_date * 1000))}</span>
      </a>
    </li>
  `}).join("\n")}
  `;
});

apiGet('/feeds/:feed_id/items/:item_id', async (req, res) => {
  const feed_id = req.params.feed_id;
  const item_id = req.params.item_id;

  query.readItem.run(Date.now(), item_id);
  return query.getFeedItemInfo.get(item_id, feed_id);

}, (output: any): string => {
  return `
  <div id="meta">
  <b>Title:</b> ${output.title}<br>
  <b>Link:</b> <a href="${output.link}" target="_blank">${output.link}</a><br>
  <b>Posted:</b> ${new Date(output.pub_date * 1000)}<brj
  <b>Read: </b> ${new Date(output.read_at || undefined)}
  </div>
  <div class="scrollable" id="feed-content">${output.content}</div>
  `
  return output.content;
});

async function periodicIngest() {
  await ingest();
  setTimeout(periodicIngest, 1000 * 60 * 10);
}

periodicIngest();

app.listen(process.env.PORT || 8000, () => {
  console.log('Listening on port', process.env.PORT || 8000);
});
